@@ -13,11 +13,12 @@ import pathlib
 import os
 import sys
 from contextlib import contextmanager
-from typing import List, Iterable, Optional
+from typing import List, Iterable, Optional, Tuple
 
 import logging
 import daiquiri
 
+import repomate
 from repomate import command
 from repomate import github_api
 from repomate import git
@@ -25,13 +26,14 @@ from repomate import util
 from repomate import tuples
 from repomate import exception
 from repomate import config
-from repomate import APIWrapper
-from repomate import hookspec
+
+import repomate_plug as plug
 
 daiquiri.setup(
     level=logging.INFO,
     outputs=(
         daiquiri.output.Stream(
+            sys.stdout,
             formatter=daiquiri.formatter.ColorFormatter(
                 fmt="[%(levelname)s] %(message)s")),
         daiquiri.output.File(
@@ -43,14 +45,33 @@ daiquiri.setup(
 
 LOGGER = daiquiri.getLogger(__file__)
 SUB = 'subparser'
+
+# Any new subparser mus tbe added to the PARSER_NAMES tuple!
 SETUP_PARSER = 'setup'
 UPDATE_PARSER = 'update'
 CLONE_PARSER = 'clone'
 MIGRATE_PARSER = 'migrate'
 ADD_TO_TEAMS_PARSER = 'add-to-teams'
-OPEN_ISSUE_PARSER = 'open-issue'
-CLOSE_ISSUE_PARSER = 'close-issue'
+OPEN_ISSUE_PARSER = 'open-issues'
+CLOSE_ISSUE_PARSER = 'close-issues'
+LIST_ISSUES_PARSER = 'list-issues'
 VERIFY_PARSER = 'verify-settings'
+ASSIGN_REVIEWS_PARSER = 'assign-peer-reviews'
+PURGE_REVIEW_TEAMS_PARSER = 'purge-peer-review-teams'
+
+PARSER_NAMES = (
+    SETUP_PARSER,
+    UPDATE_PARSER,
+    CLONE_PARSER,
+    MIGRATE_PARSER,
+    ADD_TO_TEAMS_PARSER,
+    OPEN_ISSUE_PARSER,
+    CLOSE_ISSUE_PARSER,
+    LIST_ISSUES_PARSER,
+    VERIFY_PARSER,
+    ASSIGN_REVIEWS_PARSER,
+    PURGE_REVIEW_TEAMS_PARSER,
+)
 
 
 def parse_args(sys_args: Iterable[str]
@@ -74,10 +95,11 @@ def parse_args(sys_args: Iterable[str]
             org_name=args.org_name,
             github_base_url=args.github_base_url,
             user=args.user,
+            traceback=args.traceback,
         ), None  # only here is return None for api allowed
     elif getattr(args, SUB) == CLONE_PARSER:
         # only if clone is chosen should plugins be able to hook in
-        hookspec.pm.hook.parse_args(args=args)
+        plug.manager.hook.parse_args(args=args)
 
     api = _connect_to_api(args.github_base_url, git.OAUTH_TOKEN, args.org_name)
 
@@ -105,6 +127,11 @@ def parse_args(sys_args: Iterable[str]
         issue=util.read_issue(args.issue)
         if 'issue' in args and args.issue else None,
         title_regex=args.title_regex if 'title_regex' in args else None,
+        traceback=args.traceback,
+        state=args.state if 'state' in args else None,
+        show_body=args.show_body if 'show_body' in args else None,
+        author=args.author if 'author' in args else None,
+        num_reviews=args.num_reviews if 'num_reviews' in args else None,
     )
 
     return parsed_args, api
@@ -149,12 +176,69 @@ def dispatch_command(args: tuples.Args, api: github_api.GitHubAPI):
         with _sys_exit_on_expected_error():
             command.clone_repos(args.master_repo_names, args.students, api)
     elif args.subparser == VERIFY_PARSER:
-        APIWrapper.verify_settings(args.user, args.org_name,
-                                   args.github_base_url, git.OAUTH_TOKEN)
+        github_api.GitHubAPI.verify_settings(
+            args.user, args.org_name, args.github_base_url, git.OAUTH_TOKEN)
+    elif args.subparser == LIST_ISSUES_PARSER:
+        with _sys_exit_on_expected_error():
+            command.list_issues(
+                args.master_repo_names,
+                args.students,
+                api,
+                state=args.state,
+                title_regex=args.title_regex or "",
+                show_body=args.show_body,
+                author=args.author)
+    elif args.subparser == ASSIGN_REVIEWS_PARSER:
+        with _sys_exit_on_expected_error():
+            command.assign_peer_reviews(args.master_repo_names, args.students,
+                                        args.num_reviews, args.issue, api)
+    elif args.subparser == PURGE_REVIEW_TEAMS_PARSER:
+        with _sys_exit_on_expected_error():
+            command.purge_review_teams(args.master_repo_names, args.students,
+                                       api)
     else:
         raise exception.ParseError(
-            "Illegal value for subparser: {}. This is a bug, please open an issue.".
-            format(args.subparser))
+            "Illegal value for subparser: {}. This is a bug, please open an issue."
+            .format(args.subparser))
+
+
+def _add_peer_review_parsers(base_parsers, subparsers):
+    assign_parser = subparsers.add_parser(
+        ASSIGN_REVIEWS_PARSER,
+        description=
+        "For each student repo, create a review team with pull access "
+        "named <student>-<master_repo_name>-review and randomly assign other "
+        "students to it. All students are assigned to the same amount of "
+        "review teams, as specified by `--num-reviews`. Note that "
+        "`--num-reviews` must be strictly less than the amount of students.",
+        help="Randomly assign students to peer review each others' repos.",
+        parents=base_parsers)
+    assign_parser.add_argument(
+        '-n',
+        '--num-reviews',
+        metavar='N',
+        help="Assign each student to review n repos (consequently, each repo "
+        "is reviewed by n students). n must be strictly smaller than the "
+        "amount of students.",
+        type=int,
+        default=1,
+    )
+    assign_parser.add_argument(
+        '-i',
+        '--issue',
+        help=
+        "Path to an issue to open in student repos. If specified, this issue "
+        "will be opened in each student repo, and the body will be "
+        "prepended with user mentions of all students assigned to review the "
+        "repo. NOTE: The first line is assumed to be the title.",
+        type=str,
+    )
+    purge_team_parser = subparsers.add_parser(
+        PURGE_REVIEW_TEAMS_PARSER,
+        description="Remove review teams assigned with `assign-peer-reviews`",
+        help="Remove all review teams associated with the specified "
+        "students and master repos.",
+        parents=base_parsers)
 
 
 def _add_issue_parsers(base_parsers, subparsers):
@@ -194,6 +278,55 @@ def _add_issue_parsers(base_parsers, subparsers):
         type=str,
         required=True)
 
+    list_parser = subparsers.add_parser(
+        LIST_ISSUES_PARSER,
+        description="List issues in student repos.",
+        help="List issues in student repos.",
+        parents=base_parsers,
+    )
+    list_parser.add_argument(
+        '-r',
+        '--title-regex',
+        help=
+        ("Regex to match against titles. Only issues matching this regex will "
+         "be listed."),
+    )
+    list_parser.add_argument(
+        '-b',
+        '--show-body',
+        action='store_true',
+        help="Show the body of the issue, alongside the default info.")
+    list_parser.add_argument(
+        '-a',
+        '--author',
+        help="Only show issues by this author (GitHub username).",
+        type=str,
+        default=None,
+    )
+    state = list_parser.add_mutually_exclusive_group()
+    state.add_argument(
+        '--open',
+        help="List open issues (default).",
+        action='store_const',
+        dest='state',
+        const='open',
+    )
+    state.add_argument(
+        '--closed',
+        help="List closed issues.",
+        action='store_const',
+        dest='state',
+        const='closed',
+    )
+    state.add_argument(
+        '--all',
+        help="List all issues (open and closed).",
+        action='store_const',
+        dest='state',
+        const='all',
+    )
+    list_parser.set_defaults(state='open')
+
 
 def _create_parser():
     """Create the parser."""
@@ -201,6 +334,12 @@ def _create_parser():
     parser = argparse.ArgumentParser(
         prog='repomate',
         description='A CLI tool for administrating student repositories.')
+    parser.add_argument(
+        '-v',
+        '--version',
+        help="Display version info",
+        action='version',
+        version="{} v{}".format(repomate.__package__, repomate.__version__))
     _add_subparsers(parser)
     return parser
 
@@ -294,7 +433,7 @@ def _add_subparsers(parser):
         description="Clone student repos asynchronously in bulk.",
         parents=[base_student_parser, repo_name_parser])
 
-    hookspec.pm.hook.clone_parser_hook(clone_parser=clone)
+    plug.manager.hook.clone_parser_hook(clone_parser=clone)
 
     add_to_teams = subparsers.add_parser(
         ADD_TO_TEAMS_PARSER,
@@ -311,6 +450,8 @@ def _add_subparsers(parser):
         parents=[base_student_parser, base_parser])
 
     _add_issue_parsers([base_student_parser, repo_name_parser], subparsers)
+    _add_peer_review_parsers([base_student_parser, repo_name_parser],
+                             subparsers)
 
     verify = subparsers.add_parser(
         VERIFY_PARSER,
@@ -349,6 +490,13 @@ def _create_base_parsers():
         required=is_required('github_base_url'),
         default=default('github_base_url'))
 
+    base_parser.add_argument(
+        '-tb',
+        '--traceback',
+        help=
+        "Let any exceptions propagate up so that the full traceback can be viewed.",
+        action='store_true',
+    )
     # base parser for when student lists are involved
     base_student_parser = argparse.ArgumentParser(add_help=False)
     students = base_student_parser.add_mutually_exclusive_group(
@@ -389,8 +537,8 @@ def _sys_exit_on_expected_error():
         yield
     except exception.PushFailedError as exc:
         LOGGER.error(
-            "There was an error pushing to {}. Verify that your token has adequate access.".
-            format(exc.url))
+            "There was an error pushing to {}. Verify that your token has adequate access."
+            .format(exc.url))
         sys.exit(1)
     except exception.CloneFailedError as exc:
         LOGGER.error(
@@ -447,8 +595,8 @@ def _connect_to_api(github_base_url: str, token: str,
     except exception.NotFoundError:
         # more informative message
         raise exception.NotFoundError(
-            "either organization {} could not be found, or the base url '{}' is incorrect".
-            format(org_name, github_base_url))
+            "either organization {} could not be found, or the base url '{}' is incorrect"
+            .format(org_name, github_base_url))
     return api
 
 
@@ -481,3 +629,31 @@ def _repo_names_to_urls(repo_names: Iterable[str],
     ]
 
     return non_local_urls + local_uris
+
+
+def parse_plugins(sys_args: Tuple[str]):
+    """Parse all plugin arguments.
+    
+    Args:
+        sys_args: Command line arguments.
+    """
+    parser = argparse.ArgumentParser(
+        prog='repomate',
+        description='A CLI tool for administrating student repositories.')
+    mutex_grp = parser.add_mutually_exclusive_group(required=True)
+    mutex_grp.add_argument(
+        '-p',
+        '--plug',
+        help="Specify the name of a plugin to use.",
+        type=str,
+        action='append',
+    )
+    mutex_grp.add_argument(
+        '--no-plugins',
+        help="Disable plugins.",
+        action='store_true',
+    )
+
+    args = parser.parse_args(sys_args)
+
+    return args
