@@ -13,22 +13,27 @@ program.
 .. moduleauthor:: Simon Larsén
 """
 
-import shutil
 import os
 import tempfile
-from typing import Iterable, List, Optional
+import random
+from typing import Iterable, List, Optional, Tuple, Generator
 from collections import namedtuple
+
 import daiquiri
+
+from repomate_plug import Status
+
 from repomate import git
 from repomate import github_api
 from repomate import util
 from repomate import tuples
 from repomate import exception
-from repomate import hookspec
 from repomate.github_api import GitHubAPI
 from repomate.tuples import Team
 from repomate.git import Push
 
+import repomate_plug as plug
+
 LOGGER = daiquiri.getLogger(__file__)
 
 MASTER_TEAM = 'master_repos'
@@ -197,7 +202,106 @@ def _open_issue_by_urls(repo_urls: Iterable[str], issue: tuples.Issue,
     api: A GitHubAPI to use.
     """
     repo_names = [util.repo_name(url) for url in repo_urls]
-    api.open_issue(issue, repo_names)
+    api.open_issue(issue.title, issue.body, repo_names)
+
+
+def list_issues(master_repo_names: Iterable[str],
+                students: Iterable[str],
+                api: GitHubAPI,
+                state: str = 'open',
+                title_regex: str = "",
+                show_body: bool = False,
+                author: Optional[str] = None) -> None:
+    """List all issues in the specified repos.
+
+    Args:
+        master_repo_names: Names of master repositories.
+        students: An iterable of student GitHub usernames.
+        state: state of the repo (open or closed). Defaults to 'open'.
+        api: A GitHubAPI instance used to interface with the GitHub instance.
+        title_regex: If specified, only issues with titles matching the regex
+        are displayed. Defaults to the empty string (which matches everything).
+        show_body: If True, the body of the issue is displayed along with the
+        default info.
+        author: Only show issues by this author.
+    """
+    util.validate_types(api=(api, GitHubAPI))
+    util.validate_non_empty(
+        master_repo_names=master_repo_names, students=students)
+
+    repo_names = util.generate_repo_names(students, master_repo_names)
+    max_repo_name_length = max(map(len, repo_names))
+
+    issues_per_repo = api.get_issues(repo_names, state, title_regex)
+
+    if author:
+        issues_per_repo = ((repo_name, (issue for issue in issues
+                                        if issue.user.login == author))
+                           for repo_name, issues in issues_per_repo)
+
+    _log_repo_issues(issues_per_repo, show_body, max_repo_name_length + 6)
+
+
+def _log_repo_issues(
+        issues_per_repo: Tuple[str, Generator[tuples.Issue, None, None]],
+        show_body: bool, title_alignment: int) -> None:
+    """Log repo issues.
+    
+    Args:
+        issues_per_repo: (repo_name, issue generator) pairs
+        show_body: Include the body of the issue in the output.
+        title_alignment: Where the issue title should start counting from the
+        start of the line.
+    """
+    from colored import bg, fg, style
+    even = True
+    for repo_name, issues in issues_per_repo:
+        issues = list(issues)
+
+        if not issues:
+            LOGGER.warning("{}: No matching issues".format(repo_name))
+
+        for issue in issues:
+            color = (bg('grey_30') if even else bg('grey_15')) + fg('white')
+            even = not even  # cycle color
+            adjusted_alignment = title_alignment + len(
+                color)  # color takes character space
+
+            id_ = "{}{}/#{}:".format(color, repo_name,
+                                     issue.number).ljust(adjusted_alignment)
+            out = "{}{}{}{}created {!s} by {}".format(
+                id_, issue.title, style.RESET, " ", issue.created_at,
+                issue.user.login)
+            if show_body:
+                out += os.linesep * 2 + _limit_line_length(issue.body)
+            LOGGER.info(out)
+
+
+def _limit_line_length(s: str, max_line_length: int = 100) -> str:
+    """Return the input string with lines no longer than max_line_length.
+
+    Args:
+        s: Any string.
+        max_line_length: Maximum allowed line length.
+    Returns:
+        the input string with lines no longer than max_line_length.
+    """
+    lines = s.split(os.linesep)
+    out = ""
+    for line in lines:
+        cur = 0
+        while len(line) - cur > max_line_length:
+            # find ws closest to the line length
+            idx = line.rfind(' ', cur, max_line_length + cur)
+            idx = max_line_length + cur if idx <= 0 else idx
+            if line[idx] == ' ':
+                out += line[cur:idx]
+            else:
+                out += line[cur:idx + 1]
+            out += os.linesep
+            cur = idx + 1
+        out += line[cur:cur + max_line_length] + os.linesep
+    return out
 
 
 def open_issue(issue: tuples.Issue, master_repo_names: Iterable[str],
@@ -216,7 +320,7 @@ def open_issue(issue: tuples.Issue, master_repo_names: Iterable[str],
 
     repo_names = util.generate_repo_names(students, master_repo_names)
 
-    api.open_issue(issue, repo_names)
+    api.open_issue(issue.title, issue.body, repo_names)
 
 
 def close_issue(title_regex: str, master_repo_names: Iterable[str],
@@ -259,19 +363,20 @@ def clone_repos(master_repo_names: Iterable[str], students: Iterable[str],
     LOGGER.info("cloning into student repos ...")
     git.clone(repo_urls)
 
-    if hookspec.pm.get_plugins():
-        _execute_post_clone_hooks(repo_names)
+    if len(plug.manager.
+           get_plugins()) > 1:  # something else than the default loaded
+        _execute_post_clone_hooks(repo_names, api)
 
 
-def _execute_post_clone_hooks(repo_names: List[str]):
+def _execute_post_clone_hooks(repo_names: List[str], api: GitHubAPI):
     LOGGER.info("executing post clone hooks on repos")
     local_repos = [name for name in os.listdir() if name in repo_names]
 
     results = {}
     for repo_name in local_repos:
         LOGGER.info("executing post clone hooks on {}".format(repo_name))
-        res = hookspec.pm.hook.act_on_cloned_repo(
-            path=os.path.abspath(repo_name))
+        res = plug.manager.hook.act_on_cloned_repo(
+            path=os.path.abspath(repo_name), api=api)
         results[repo_name] = res
     LOGGER.info(_format_hook_results_output(results))
 
@@ -309,19 +414,80 @@ def migrate_repos(master_repo_urls: Iterable[str], user: str,
         _clone_all(master_repo_urls, cwd=tmpdir)
         repo_urls = api.create_repos(infos)
 
-        git.push(
-            [
-                git.Push(
-                    local_path=os.path.join(tmpdir, info.name),
-                    repo_url=repo_url,
-                    branch='master')
-                for repo_url, info in zip(repo_urls, infos)
-            ],
-            user=user)
+        git.push([
+            git.Push(
+                local_path=os.path.join(tmpdir, info.name),
+                repo_url=repo_url,
+                branch='master') for repo_url, info in zip(repo_urls, infos)
+        ],
+                 user=user)
 
     LOGGER.info("done!")
 
 
+def assign_peer_reviews(
+        master_repo_names: Iterable[str],
+        students: Iterable[str],
+        num_reviews: int,
+        issue: Optional[tuples.Issue],
+        api: GitHubAPI,
+) -> None:
+    """Assign peer reviewers among the students to each student repo. Each
+    student is assigned to review num_reviews repos, and consequently, each
+    repo gets reviewed by num_reviews reviewers.
+
+    In practice, each student repo has a review team generated (called
+    <student-repo-name>-review), to which num_reviews _other_ students are
+    assigned. The team itself is given pull-access to the student repo, so
+    that reviewers can view code and open issues, but cannot modify the
+    contents of the repo.
+
+    Args:
+        master_repo_names: Names of master repos.
+        students: An iterable of student GitHub usernames.
+        num_reviews: Amount of reviews each student should perform
+            (consequently, the amount of reviews of each repo)
+        api: A GitHubAPI instance used to interface with the GitHub instance.
+    """
+    util.validate_types(
+        api=(api, GitHubAPI), issue=(issue, (tuples.Issue, type(None))))
+    util.validate_non_empty(
+        master_repo_names=master_repo_names, students=students)
+
+    for master_name in master_repo_names:
+        peer_review_allocations = plug.manager.hook.generate_review_allocations(
+            master_repo_name=master_name,
+            students=students,
+            num_reviews=num_reviews,
+            review_team_name_function=util.generate_review_team_name)
+        api.ensure_teams_and_members(
+            peer_review_allocations, permission='pull')
+        api.add_repos_to_review_teams({
+            util.generate_review_team_name(student, master_name):
+            [util.generate_repo_name(student, master_name)]
+            for student in students
+        },
+                                      issue=issue)
+
+
+def purge_review_teams(master_repo_names: Iterable[str],
+                       students: Iterable[str], api: GitHubAPI) -> None:
+    """Delete all review teams associated with the given master repo names and students.
+    
+    Args:
+        master_repo_names: Names of master repos.
+        students: An iterable of student GitHub usernames.
+    """
+    util.validate_non_empty(
+        master_repo_names=master_repo_names, students=students)
+
+    review_team_names = [
+        util.generate_review_team_name(student, master_repo_name)
+        for student in students for master_repo_name in master_repo_names
+    ]
+    api.delete_teams(review_team_names)
+
+
 def _create_repo_infos(urls: Iterable[str],
                        teams: Iterable[Team]) -> List[tuples.Repo]:
     """Create Repo namedtuples for all combinations of url and team.
@@ -369,16 +535,24 @@ def _create_push_tuples(master_repo_paths: Iterable[str],
         ]
     return push_tuples
 
-def _format_hook_result(hook_result):
-    from colored import bg, style
-    if hook_result.status == "error":
-        out = bg('red')
-    elif hook_result.status == "warning":
-        out = bg('yellow')
-    else:
-        out = bg('dark_green')
 
-    out += hook_result.hook + ": " + hook_result.status + style.RESET + os.linesep
+def _format_hook_result(hook_result):
+    from colored import bg, fg, style
+    if hook_result.status == Status.ERROR:
+        out = bg('red')
+    elif hook_result.status == Status.WARNING:
+        out = bg('yellow')
+    elif hook_result.status == Status.SUCCESS:
+        out = bg('dark_green')
+    else:
+        raise ValueError(
+            "expected hook_result.status to be one of Status.ERROR, "
+            "Status.WARNING or Status.SUCCESS, but was {!r}".format(
+                hook_result.status))
+
+    out += fg(
+        'white'
+    ) + hook_result.hook + ": " + hook_result.status.name + style.RESET + os.linesep
     out += hook_result.msg
 
     return out
@@ -389,7 +563,7 @@ def _format_hook_results_output(result_mapping):
     out = ""
     for repo_name, results in result_mapping.items():
         out += "{}hook results for {}{}{}".format(
-            bg('dark_gray'), repo_name, style.RESET, os.linesep * 2)
+            bg('grey_23'), repo_name, style.RESET, os.linesep * 2)
         out += os.linesep.join([
             "{}{}".format(_format_hook_result(res), os.linesep)
             for res in results
@@ -397,5 +571,3 @@ def _format_hook_results_output(result_mapping):
         out += os.linesep * 2
 
     return out
-
-
